@@ -43,8 +43,14 @@ PROG = Path(__file__).stem
 CNFFILE = Path(os.getenv('XDG_CONFIG_HOME', '~/.config'), f'{PROG}.conf')
 CWD = Path.cwd()
 commands = []
+options = {}
+aliases_all = set()
 cnffile = None
 
+def get_editor():
+    'Return editor for this user'
+    return os.getenv('VISUAL') or os.getenv('EDITOR') or 'vi'
+
 def infer_root(path, *, dest=False):
     'Infer leading directory path'
     def _parse(path, dest):
@@ -73,6 +79,11 @@ def infer_root(path, *, dest=False):
 
 infer_root.lead = ''
 
+def infer_path(path):
+    'Get inferred path'
+    ipath = infer_root(path)
+    return None if ipath is None else Path(ipath)
+
 def doexit(args, code_or_msg=0):
     'Exit but check if final hard/soft reset is required'
     if args.reset:
@@ -96,7 +107,7 @@ def mpcmd(args, cmdstr, quiet=False):
         if args.mount_unsafe_links:
             arglist.append(f'mount -l {args.mount_unsafe_links}')
         elif args.mount:
-            arglist.append(f'mount -l {args.mount}')
+            arglist.append(f'mount {args.mount}')
 
         mpcmd.cmdtext = ' '.join(arglist)
 
@@ -113,6 +124,27 @@ def mpcmd(args, cmdstr, quiet=False):
 
 mpcmd.cmdtext = None
 
+def mpcmd_wrap(args):
+    'Extract args/options and send to device'
+    opts = options[args.cmdname]
+    arglist = [args.cmdname]
+    for opt in opts._actions:
+        arg = args.__dict__.get(opt.dest)
+        if arg is None:
+            continue
+        if opt.const:
+            if arg == opt.default:
+                continue
+            arg = None
+
+        if opt.option_strings:
+            arglist.append(opt.option_strings[-1])
+
+        if arg is not None:
+            arglist.append(' '.join(arg) if isinstance(arg, list) else arg)
+
+    mpcmd(args, ' '.join(arglist))
+
 def main():
     'Main code'
     global cnffile
@@ -122,24 +154,23 @@ def main():
             epilog=f'Type "{PROG} <command> -h" to see specific help/usage '
             ' for any of the above commands. Note you can set default '
             f'options in {CNFFILE}. '
-            f'Use "{PROG} edit" to conveniently change the file.')
+            f'Use "{PROG} config" to conveniently change the file.')
 
     # Set up main/global arguments
     opt.add_argument('-d', '--device',
-            help='serial port/device to connect to, default is "auto"')
-    opt.add_argument('-l', '--list-device-names', action='store_true',
-            help='just list out device mnemonics that can be '
-            'used for -d/--device')
+            help='serial port/device to connect to, default is "auto". '
+                 'Specify "-d list" to print out device mnemonics that '
+                 'can be used.')
     opt.add_argument('-m', '--mount',
             help='mount local directory on device before command')
     opt.add_argument('-M', '--mount-unsafe-links',
             help='mount local directory and allow external links')
-    opt.add_argument('-r', '--reset-hard', dest='reset',
-            action='store_const', const=2,
-            help='do hard reset after command')
-    opt.add_argument('-s', '--reset-soft', dest='reset',
+    opt.add_argument('-x', '--reset', dest='reset',
             action='store_const', const=1,
             help='do soft reset after command')
+    opt.add_argument('-b', '--reboot', dest='reset',
+            action='store_const', const=2,
+            help='do hard reboot after command')
     opt.add_argument('-p', '--path-to-mpremote', default='mpremote',
             help='path to mpremote program, default = "%(default)s"')
     opt.add_argument('-c', '--completion', action='store_true',
@@ -160,12 +191,18 @@ def main():
             sys.exit(f'Must define a docstring for command class "{name}".')
 
         title = desc.splitlines()[0]
-        assert title[-1] == '.', f'Title "{title}" should include full stop.'
+        if title[-1] != '.':
+            sys.exit(f'Title "{title}" should include full stop.')
 
+        # Code check to ensure we have not defined duplicate aliases
         aliases = cls.aliases if hasattr(cls, 'aliases') else []
+        for a in aliases:
+            if a in aliases_all:
+                sys.exit(f'command {name}: duplicate alias: {a}')
+            aliases_all.add(a)
 
-        cmdopt = cmd.add_parser(name, description=desc, help=title,
-                aliases=aliases)
+        options[name] = cmdopt = cmd.add_parser(name, description=desc,
+                                help=title, aliases=aliases)
 
         # Set up this commands own arguments, if it has any
         if hasattr(cls, 'init'):
@@ -197,7 +234,7 @@ def main():
         return
 
     # Just print out device names if asked
-    if args.list_device_names:
+    if args.device == 'list':
         print(DEVICE_NAMES)
         return
 
@@ -242,7 +279,7 @@ class _get(COMMAND):
             parent.mkdir(exist_ok=True, parents=True)
 
         for argsrc in args.src:
-            src = Path(infer_root(argsrc))
+            src = infer_path(argsrc)
             srcstr = src and str(src).lstrip('/')
             if srcstr:
                 if dst:
@@ -289,6 +326,27 @@ class _put(COMMAND):
             filedst = filedst.lstrip('/')
             mpcmd(args, f'{arg} {src} :{filedst}')
 
+@command
+class _copy(COMMAND):
+    'Copy one of more remote files to a directory on device.'
+    aliases = ['c']
+
+    def init(opt):
+        opt.add_argument('-f', '--file', action='store_true',
+                help='destination is file, not directory')
+        opt.add_argument('src', nargs='+',
+                help='name of source file[s] on device')
+        opt.add_argument('dst',
+                help='name of destination dir on device')
+
+    def run(args):
+        dst = Path(infer_root(args.dst, dest=True))
+
+        for argsrc in args.src:
+            src = Path(argsrc)
+            filedst = str(dst if args.file else dst / src.name).lstrip('/')
+            mpcmd(args, f'cp :{src} :{filedst}')
+
 @command
 class _ls(COMMAND):
     'List directory on device.'
@@ -348,28 +406,53 @@ class _rm(COMMAND):
                 mpcmd(args, f'rm {path}', args.quiet)
 
 @command
-class _reset(COMMAND):
-    'Soft or hard reset the device.'
-    aliases = ['x']
+class _touch(COMMAND):
+    'Touch the given file[s] on device.'
 
     def init(opt):
-        opt.add_argument('-b', '--reboot', action='store_true',
-                help='Do full reboot, i.e. a hard reset')
-        opt.add_argument('delay_ms', type=int, nargs='?',
-                help='optional delay before hard reset (millisecs)')
+        opt.add_argument('file', nargs='+', help='name of file[s]')
+
+    def run(args):
+        mpcmd_wrap(args)
+
+@command
+class _edit(COMMAND):
+    '''
+    Edit the given file[s] on device.
+
+    Copies the file from device, opens your editor on that local file,
+    then copies it back.
+    '''
+    aliases = ['e']
+
+    def init(opt):
+        opt.add_argument('file', nargs='+', help='name of file[s]')
+
+    def run(args):
+        mpcmd_wrap(args)
+
+@command
+class _reset(COMMAND):
+    'Soft reset the device.'
+    aliases = ['x']
 
     def run(args):
         args.reset = None
-        if args.reboot:
-            arg = 'reset'
-            if args.delay_ms:
-                arg += f' {args.delay_ms}'
-        else:
-            if args.delay_ms:
-                sys.exit('Delay can only be used with hard reset.')
-            arg = 'soft-reset'
+        mpcmd(args, 'soft-reset')
 
-        mpcmd(args, arg)
+@command
+class _reboot(COMMAND):
+    'Hard reboot the device.'
+    aliases = ['b']
+
+    def init(opt):
+        opt.add_argument('delay_ms', type=int, nargs='?',
+                help='optional delay before reboot (millisecs)')
+
+    def run(args):
+        args.reset = None
+        arg = f' {args.delay_ms}' if args.delay_ms else ''
+        mpcmd(args, 'reset' + arg)
 
 @command
 class _repl(COMMAND):
@@ -385,18 +468,11 @@ class _repl(COMMAND):
                 help='file to inject at the REPL when Ctrl-K is pressed')
 
     def run(args):
-        arglist = ['repl']
-        for opt in ('capture', 'inject_code', 'inject_file'):
-            arg = args.__dict__.get(opt)
-            if arg:
-                opt = opt.replace('_', '-')
-                arglist.append(f'--{opt} "{arg}"')
-
-        mpcmd(args, ' '.join(arglist))
+        mpcmd_wrap(args)
 
 @command
 class _list(COMMAND):
-    'List currently available devices.'
+    'List currently connected devices.'
     aliases = ['l', 'devs']
 
     def run(args):
@@ -438,6 +514,27 @@ class _exec(COMMAND):
         for string in args.string:
             mpcmd(args, f'exec "{string}"')
 
+@command
+class _mip(COMMAND):
+    'Run mip to install packages on device.'
+    aliases = ['m']
+
+    def init(opt):
+        opt.add_argument('-n', '--no-mpy', action='store_true',
+                help='download .py files, not compiled .mpy files')
+        opt.add_argument('-t', '--target',
+                help='destination directory on device')
+        opt.add_argument('-i', '--index',
+                help='package index to use, default="micropython-lib"')
+        opt.add_argument('command',
+                help='mip command, e.g. "install"')
+        opt.add_argument('package', nargs='+',
+                help='package specifications, e.g. "name", "name@version", '
+                         '"github.org/repo", "github.org/repo@branch"')
+
+    def run(args):
+        mpcmd_wrap(args)
+
 @command
 class _bootloader(COMMAND):
     'Enter bootloader on device.'
@@ -455,12 +552,12 @@ class _version(COMMAND):
     'Show version of mpremote tool.'
 
 @command
-class _edit(COMMAND):
-    doc = f'Open the {PROG} configuration file with your $VISUAL editor.'
+class _config(COMMAND):
+    doc = f'Open the {PROG} configuration file with your editor.'
+    aliases = ['cf']
 
     def run(args):
-        editor = os.getenv('VISUAL') or os.getenv('EDITOR') or 'vi'
-        subprocess.run(f'{editor} {cnffile}'.split())
+        subprocess.run(f'{get_editor()} {cnffile}'.split())
 
 if __name__ == '__main__':
     main()
